﻿using System;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Collections.Generic;
using HeuristicLab.Algorithms.GeneticAlgorithm;
using System.Threading.Tasks;
using System.Diagnostics;
using System.Reflection;

namespace HeuristicLab.Problems.MetaOptimization {
  /// <summary>
  /// A base class for operators which evaluate TSP solutions.
  /// </summary>
  [Item("ParameterConfigurationEvaluator", "A base class for operators which evaluate Meta Optimization solutions.")]
  [StorableClass]
  public class ParameterConfigurationEvaluator : SingleSuccessorOperator, IParameterConfigurationEvaluator {
    private const double PenaltyQuality = 100000.0; // todo: use something better

    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<EngineAlgorithm> AlgorithmParameter {
      get { return (ILookupParameter<EngineAlgorithm>)Parameters[MetaOptimizationProblem.AlgorithmTypeParameterName]; }
    }
    public ILookupParameter<IItemList<ISingleObjectiveProblem>> ProblemsParameter {
      get { return (ILookupParameter<IItemList<ISingleObjectiveProblem>>)Parameters[MetaOptimizationProblem.ProblemsParameterName]; }
    }
    public ILookupParameter<ParameterConfigurationTree> ParameterConfigurationParameter {
      get { return (ILookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }
    public LookupParameter<IntValue> RepetitionsParameter {
      get { return (LookupParameter<IntValue>)Parameters[MetaOptimizationProblem.RepetitionsParameterName]; }
    }

    public LookupParameter<DoubleArray> ProblemQualityReferencesParameter {
      get { return (LookupParameter<DoubleArray>)Parameters["ProblemQualityReferences"]; }
    }

    public IntValue Repetitions {
      get { return RepetitionsParameter.ActualValue; }
    }

    public ParameterConfigurationEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The evaluated quality of the ParameterVector."));
      Parameters.Add(new LookupParameter<EngineAlgorithm>(MetaOptimizationProblem.AlgorithmTypeParameterName, "Missing description."));
      Parameters.Add(new LookupParameter<IItemList<ISingleObjectiveProblem>>(MetaOptimizationProblem.ProblemsParameterName, "Missing description."));
      Parameters.Add(new LookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", "Missing description."));
      Parameters.Add(new LookupParameter<IntValue>(MetaOptimizationProblem.RepetitionsParameterName, "Number of evaluations on one problem."));
      Parameters.Add(new LookupParameter<DoubleArray>("ProblemQualityReferences", ""));
    }

    [StorableConstructor]
    protected ParameterConfigurationEvaluator(bool deserializing) : base(deserializing) { }
    protected ParameterConfigurationEvaluator(ParameterConfigurationEvaluator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ParameterConfigurationEvaluator(this, cloner);
    }

    public override IOperation Apply() {
      EngineAlgorithm algorithm = (EngineAlgorithm)AlgorithmParameter.ActualValue.Clone();
      IItemList<ISingleObjectiveProblem> problems = ProblemsParameter.ActualValue;
      ParameterConfigurationTree parameterConfiguration = ParameterConfigurationParameter.ActualValue;

      double[] referenceQualities;
      if (ProblemQualityReferencesParameter.ActualValue == null) {
        // this is generation zero. no reference qualities for normalization have been calculated yet.
        referenceQualities = new double[problems.Count];
        for (int i = 0; i < referenceQualities.Length; i++) {
          referenceQualities[i] = 1;
        }
      } else {
        referenceQualities = ProblemQualityReferencesParameter.ActualValue.ToArray();
      }

      // set parameters
      parameterConfiguration.Parameterize(algorithm);
      algorithm.StoreAlgorithmInEachRun = false;

      List<List<double>> qualities = new List<List<double>>();
      List<List<TimeSpan>> executionTimes = new List<List<TimeSpan>>();
      algorithm.Engine = new SequentialEngine.SequentialEngine();
      algorithm.Prepare(true);

      foreach (ISingleObjectiveProblem problem in problems) {
        algorithm.Problem = (IProblem)problem.Clone();
        var problemQualities = new List<double>();
        var problemExecutionTimes = new List<TimeSpan>();

        for (int i = 0; i < Repetitions.Value; i++) {
          algorithm.Prepare();

          AlgorithmExecutor executor = new AlgorithmExecutor(algorithm);
          executor.StartSync();

          if (algorithm.ExecutionState == ExecutionState.Paused) {
            // this parametercombination was bad. set penalty for this solution
            problemQualities.Add(PenaltyQuality); // todo: respect Maximization; problem: this messes up average value; solution: use currently worst quality*2
            problemExecutionTimes.Add(algorithm.ExecutionTime);
          } else {
            problemQualities.Add(((DoubleValue)algorithm.Results["BestQuality"].Value).Value);
            problemExecutionTimes.Add(algorithm.ExecutionTime);

            // parameters will be stored in ParameterConfigurationTree anyway. they would be redundant in runs
            algorithm.Runs.Last().Parameters.Clear();
            // but keep the problem, since this differs in runs
            algorithm.Runs.Last().Parameters.Add("Problem", problem);
          }
        }
        qualities.Add(problemQualities);
        executionTimes.Add(problemExecutionTimes);
      }
      algorithm.Prepare();

      parameterConfiguration.AverageExecutionTimes = new ItemList<TimeSpanValue>(executionTimes.Select(t => new TimeSpanValue(TimeSpan.FromMilliseconds(t.Average(ts => ts.TotalMilliseconds)))));
      parameterConfiguration.Repetitions = (IntValue)Repetitions.Clone();
      parameterConfiguration.BestQualities = new DoubleArray(qualities.Select(q => q.Min()).ToArray()); // todo: respect Maximization:true/false
      parameterConfiguration.AverageQualities = new DoubleArray(qualities.Select(q => q.Average()).ToArray());
      parameterConfiguration.WorstQualities = new DoubleArray(qualities.Select(q => q.Max()).ToArray()); // todo: respect Maximization:true/false
      parameterConfiguration.QualityVariances = new DoubleArray(qualities.Select(q => q.Variance()).ToArray());
      parameterConfiguration.QualityStandardDeviations = new DoubleArray(qualities.Select(q => q.StandardDeviation()).ToArray());
      parameterConfiguration.Runs = (RunCollection)algorithm.Runs.Clone();

      this.QualityParameter.ActualValue = new DoubleValue(NormalizeQualities(parameterConfiguration, referenceQualities));

      return base.Apply();
    }

    public static double NormalizeQualities(ParameterConfigurationTree parameterConfigurationTree, double[] referenceQualities) {
      double[] qualitiesNormalized = new double[referenceQualities.Length];
      for (int i = 0; i < referenceQualities.Length; i++) {
        qualitiesNormalized[i] = parameterConfigurationTree.AverageQualities[i] / referenceQualities[i];
      }
      parameterConfigurationTree.QualitiesNormalized = new DoubleArray(qualitiesNormalized);
      parameterConfigurationTree.AverageQualityNormalized = new DoubleValue(qualitiesNormalized.Average());
      return parameterConfigurationTree.AverageQualityNormalized.Value;
    }

    public static double Variance(IEnumerable<double> source) {
      double avg = source.Average();
      double d = source.Aggregate(0.0, (total, next) => total += Math.Pow(next - avg, 2));
      return d / (source.Count() - 1);
    }

    public static double StandardDeviation(IEnumerable<double> source) {
      return Math.Sqrt(source.Variance());
    }
  }
}
